%% Clear Workspace and Close figures
clear; close all; clc;

%% Intialize Laplace variable
s = zpk('s');

addpath('./mat/');

% Initialization
% Let's load the previously defined parameters for the model.


load('parameters.mat');

bode_opts = bodeoptions;
bode_opts.FreqUnits = 'Hz';
bode_opts.MagUnits  = 'abs';
bode_opts.MagScale  = 'log';
bode_opts.Grid      = 'on';
bode_opts.PhaseVisible = 'off';
bode_opts.Title.FontSize     = 10;
bode_opts.XLabel.FontSize    = 10;
bode_opts.YLabel.FontSize    = 10;
bode_opts.TickLabel.FontSize = 10;

open rotating_frame.slx



% First we define the parameters that must be defined in order to run the Simscape simulation.

w = 2*pi; % Rotation speed [rad/s]

theta_e = 0; % Static measurement error on the angle theta [rad]

m = 5; % mass of the sample [kg]

mTuv = 30;% Mass of the moving part of the Tuv stage [kg]
kTuv = 1e8; % Stiffness of the Tuv stage [N/m]
cTuv = 0; % Damping of the Tuv stage [N/(m/s)]



% Then, we defined parameters that will be used in the following analysis.

mlight = 5; % Mass for light sample [kg]
mheavy = 55; % Mass for heavy sample [kg]

wlight = 2*pi; % Max rot. speed for light sample [rad/s]
wheavy = 2*pi/60; % Max rot. speed for heavy sample [rad/s]

kvc = 1e3; % Voice Coil Stiffness [N/m]
kpz = 1e8; % Piezo Stiffness [N/m]

d = 0.01; % Maximum excentricity from rotational axis [m]

freqs = logspace(-2, 3, 1000); % Frequency vector for analysis [Hz]

% Identification in the rotating referenced frame
% We initialize the inputs and outputs of the system to identify:
% - Inputs: $f_u$ and $f_v$
% - Outputs: $d_u$ and $d_v$


%% Options for Linearized
options = linearizeOptions;
options.SampleTime = 0;

%% Name of the Simulink File
mdl = 'rotating_frame';

%% Input/Output definition
io(1) = linio([mdl, '/fu'], 1, 'input');
io(2) = linio([mdl, '/fv'], 1, 'input');

io(3) = linio([mdl, '/du'], 1, 'output');
io(4) = linio([mdl, '/dv'], 1, 'output');



% We start we identify the transfer functions at high speed with the light sample.

w = wlight; % Rotation speed [rad/s]
m = mlight; % mass of the sample [kg]

kTuv = kpz;
Gpz_light = linearize(mdl, io, 0.1);
Gpz_light.InputName  = {'Fu', 'Fv'};
Gpz_light.OutputName = {'Du', 'Dv'};

kTuv = kvc;
Gvc_light = linearize(mdl, io, 0.1);
Gvc_light.InputName  = {'Fu', 'Fv'};
Gvc_light.OutputName = {'Du', 'Dv'};



% Then we identify the system with an heavy mass and low speed.

w = wheavy; % Rotation speed [rad/s]
m = mheavy; % mass of the sample [kg]

kTuv = kpz;
Gpz_heavy = linearize(mdl, io, 0.1);
Gpz_heavy.InputName  = {'Fu', 'Fv'};
Gpz_heavy.OutputName = {'Du', 'Dv'};

kTuv = kvc;
Gvc_heavy = linearize(mdl, io, 0.1);
Gvc_heavy.InputName  = {'Fu', 'Fv'};
Gvc_heavy.OutputName = {'Du', 'Dv'};

% Coupling ratio between $f_{uv}$ and $d_{uv}$
% In order to validate the equations written, we can compute the coupling ratio using the simscape model and compare with the equations.

% From the previous identification, we plot the coupling ratio in both case (figure [[fig:coupling_ratio_light_heavy]]).

% We obtain the same result than the analytical case (figures [[fig:coupling_light]] and [[fig:coupling_heavy]]).

figure;
hold on;
plot(freqs, abs(squeeze(freqresp(Gvc_light('Du', 'Fu'), freqs, 'Hz')))./abs(squeeze(freqresp(Gvc_light('Dv', 'Fu'), freqs, 'Hz'))));
plot(freqs, abs(squeeze(freqresp(Gpz_light('Du', 'Fu'), freqs, 'Hz')))./abs(squeeze(freqresp(Gpz_light('Dv', 'Fu'), freqs, 'Hz'))));
set(gca,'ColorOrderIndex',1);
plot(freqs, abs(squeeze(freqresp(Gvc_heavy('Du', 'Fu'), freqs, 'Hz')))./abs(squeeze(freqresp(Gvc_heavy('Dv', 'Fu'), freqs, 'Hz'))), '--');
plot(freqs, abs(squeeze(freqresp(Gpz_heavy('Du', 'Fu'), freqs, 'Hz')))./abs(squeeze(freqresp(Gpz_heavy('Dv', 'Fu'), freqs, 'Hz'))), '--');
hold off;
xlim([freqs(1), freqs(end)]);
set(gca, 'XScale', 'log'); set(gca, 'YScale', 'log');
xlabel('Frequency [Hz]'); ylabel('Coupling ratio');
legend({'light - VC', 'light - PZ', 'heavy - VC',  'heavy - PZ'}, 'Location', 'northeast')

% Transfer function from force to force sensor (IFF plant)

%% Name of the Simulink File
mdl = 'rotating_frame';

%% Input/Output definition
io(1) = linio([mdl, '/fu'], 1, 'input');
io(2) = linio([mdl, '/fv'], 1, 'input');

io(3) = linio([mdl, '/fum'], 1, 'output');
io(4) = linio([mdl, '/fvm'], 1, 'output');



% We start we identify the transfer functions at high speed with the light sample.

w = wlight; % Rotation speed [rad/s]
m = mlight; % mass of the sample [kg]

kTuv = kpz;
Gpz_light = linearize(mdl, io, 0.1);
Gpz_light.InputName  = {'Fu', 'Fv'};
Gpz_light.OutputName = {'Fum', 'Fvm'};

kTuv = kvc;
Gvc_light = linearize(mdl, io, 0.1);
Gvc_light.InputName  = {'Fu', 'Fv'};
Gvc_light.OutputName = {'Fum', 'Fvm'};



% Then we identify the system with an heavy mass and low speed.

w = wheavy; % Rotation speed [rad/s]
m = mheavy; % mass of the sample [kg]

kTuv = kpz;
Gpz_heavy = linearize(mdl, io, 0.1);
Gpz_heavy.InputName  = {'Fu', 'Fv'};
Gpz_heavy.OutputName = {'Fum', 'Fvm'};

kTuv = kvc;
Gvc_heavy = linearize(mdl, io, 0.1);
Gvc_heavy.InputName  = {'Fu', 'Fv'};
Gvc_heavy.OutputName = {'Fum', 'Fvm'};

figure;
hold on;
plot(freqs, abs(squeeze(freqresp(Gvc_heavy('Fum', 'Fu'), freqs, 'Hz'))));
hold off;
xlim([freqs(1), freqs(end)]);
set(gca, 'XScale', 'log'); set(gca, 'YScale', 'log');
xlabel('Frequency [Hz]'); ylabel('Coupling ratio');
legend({'light - VC', 'light - PZ', 'heavy - VC',  'heavy - PZ'}, 'Location', 'northeast')

% Plant identification
% The goal is to study the control problems due to the coupling that appears because of the rotation.


%% Options for Linearized
options = linearizeOptions;
options.SampleTime = 0;

%% Name of the Simulink File
mdl = 'rotating_frame';

%% Input/Output definition
io(1) = linio([mdl, '/fu'], 1, 'input');
io(2) = linio([mdl, '/fv'], 1, 'input');

io(3) = linio([mdl, '/du'], 1, 'output');
io(4) = linio([mdl, '/dv'], 1, 'output');



% First, we identify the system when the rotation speed is null and then when the rotation speed is equal to 60rpm.

% The actuators are voice coil with some damping added.

% The bode plot of the system not rotating and rotating at 60rpm is shown figure [[fig:Gvc_speed]].


w = 0; % Rotation speed [rad/s]
m = mlight; % mass of the sample [kg]
kTuv = kvc;
cTuv = 0.1*sqrt(kTuv*m);

Gvc = linearize(mdl, io, 0.1);
Gvc.InputName  = {'Fu', 'Fv'};
Gvc.OutputName = {'Du', 'Dv'};

w = wlight; % Rotation speed [rad/s]
m = mlight; % mass of the sample [kg]
kTuv = kvc;
cTuv = 0.1*sqrt(kTuv*m);

Gtvc = linearize(mdl, io, 0.1);
Gtvc.InputName  = {'Fu', 'Fv'};
Gtvc.OutputName = {'Du', 'Dv'};

figure;
ax1 = subplot(2,2,1);
title('From $F_u$ to $D_u$');
hold on;
plot(freqs, abs(squeeze(freqresp(Gvc(1, 1),  freqs, 'Hz'))));
plot(freqs, abs(squeeze(freqresp(Gtvc(1, 1), freqs, 'Hz'))));
hold off;
xlim([freqs(1), freqs(end)]);
set(gca, 'XScale', 'log'); set(gca, 'YScale', 'log');
set(gca, 'XTickLabel',[]);
ylabel('Magnitude [m/N]');

ax2 = subplot(2,2,3);
hold on;
plot(freqs, 180/pi*angle(squeeze(freqresp(Gvc(1, 1),  freqs, 'Hz'))), 'DisplayName', 'Gvc - $\omega = 0$');
plot(freqs, 180/pi*angle(squeeze(freqresp(Gtvc(1, 1), freqs, 'Hz'))), 'DisplayName', 'Gvc - $\omega = 60$rpm');
hold off;
yticks(-180:90:180);
ylim([-180 180]);
xlim([freqs(1), freqs(end)]);
set(gca, 'XScale', 'log'); set(gca, 'YScale', 'lin');
xlabel('Frequency [Hz]'); ylabel('Phase [deg]');
legend('Location', 'northeast');
linkaxes([ax1,ax2],'x');

ax1 = subplot(2,2,2);
title('From $F_u$ to $D_v$');
hold on;
plot(freqs, abs(squeeze(freqresp(Gvc(1, 2),  freqs, 'Hz'))));
plot(freqs, abs(squeeze(freqresp(Gtvc(1, 2), freqs, 'Hz'))));
hold off;
xlim([freqs(1), freqs(end)]);
set(gca, 'XScale', 'log'); set(gca, 'YScale', 'log');
set(gca, 'XTickLabel',[]);

ax2 = subplot(2,2,4);
hold on;
plot(freqs, 180/pi*angle(squeeze(freqresp(Gvc(1, 2),  freqs, 'Hz'))), 'DisplayName', 'Gvc - $\omega = 0$');
plot(freqs, 180/pi*angle(squeeze(freqresp(Gtvc(1, 2), freqs, 'Hz'))), 'DisplayName', 'Gvc - $\omega = 60$rpm');
hold off;
yticks(-180:90:180);
ylim([-180 180]);
xlim([freqs(1), freqs(end)]);
set(gca, 'XScale', 'log'); set(gca, 'YScale', 'lin');
xlabel('Frequency [Hz]');
legend('Location', 'northeast');
linkaxes([ax1,ax2],'x');

% Effect of rotation speed
% We first identify the system (voice coil and light mass) for multiple rotation speed.
% Then we compute the bode plot of the system (figure [[fig:Guu_uv_ws]]).

% As the rotation frequency increases:
% - one pole goes to lower frequencies while the other goes to higher frequencies
% - one zero appears between the two poles
% - the zero disappears when $\omega > \sqrt{\frac{k}{m}}$ and the low frequency pole becomes unstable (positive real part)

% To stabilize the unstable pole, we need a control bandwidth of at least twice of frequency of the unstable pole.


ws = linspace(0, 2*pi, 5); % Rotation speed vector [rad/s]
m = mlight; % mass of the sample [kg]

kTuv = kvc;
cTuv = 0.1*sqrt(kTuv*m);

Gs_vc = {zeros(1, length(ws))};

for i = 1:length(ws)
  w = ws(i);
  Gs_vc{i} = linearize(mdl, io, 0.1);
  Gs_vc{i}.InputName  = {'Fu', 'Fv'};
  Gs_vc{i}.OutputName = {'Du', 'Dv'};
end

freqs = logspace(-2, 2, 1000);

figure;
ax1 = subplot(2,2,1);
title('$D_u/F_u$');
hold on;
for i = 1:length(ws)
  plot(freqs, abs(squeeze(freqresp(Gs_vc{i}(1, 1), freqs, 'Hz'))));
end
hold off;
xlim([freqs(1), freqs(end)]);
ylim([1e-8, 1]);
set(gca, 'XScale', 'log'); set(gca, 'YScale', 'log');
set(gca, 'XTickLabel',[]);
ylabel('Magnitude [m/N]');

ax2 = subplot(2,2,3);
hold on;
for i = 1:length(ws)
  plot(freqs, 180/pi*angle(squeeze(freqresp(Gs_vc{i}(1, 1), freqs, 'Hz'))));
end
hold off;
yticks(-180:90:180);
ylim([-180 180]);
xlim([freqs(1), freqs(end)]);
set(gca, 'XScale', 'log'); set(gca, 'YScale', 'lin');
xlabel('Frequency [Hz]'); ylabel('Phase [deg]');
linkaxes([ax1,ax2],'x');

ax1 = subplot(2,2,2);
title('$D_v/F_u$');
hold on;
for i = 1:length(ws)
  plot(freqs, abs(squeeze(freqresp(Gs_vc{i}(1, 2), freqs, 'Hz'))));
end
hold off;
xlim([freqs(1), freqs(end)]);
ylim([1e-8, 1]);
set(gca, 'XScale', 'log'); set(gca, 'YScale', 'log');
set(gca, 'XTickLabel',[]);
set(gca, 'YTickLabel',[]);

ax2 = subplot(2,2,4);
hold on;
for i = 1:length(ws)
  plot(freqs, 180/pi*angle(squeeze(freqresp(Gs_vc{i}(1, 2), freqs, 'Hz'))), 'DisplayName', sprintf('w = %.0f [rpm]', ws(i)*60/2/pi));
end
hold off;
yticks(-180:90:180);
ylim([-180 180]);
xlim([freqs(1), freqs(end)]);
set(gca, 'XScale', 'log'); set(gca, 'YScale', 'lin');
set(gca, 'YTickLabel',[]);
xlabel('Frequency [Hz]');
legend('Location', 'northeast');
linkaxes([ax1,ax2],'x');



% #+LABEL: fig:Guu_uv_ws
% #+CAPTION: Diagonal term as a function of the rotation frequency
% [[file:figs/Guu_uv_ws.png]]

% Then, we can look at the same plots for the piezoelectric actuator (figure [[fig:Guu_ws_pz]]). The effect of the rotation frequency has very little effect on the dynamics of the system to control.


ws = linspace(0, 2*pi, 5); % Rotation speed vector [rad/s]
m = mlight; % mass of the sample [kg]

kTuv = kpz;
cTuv = 0.1*sqrt(kTuv*m);

Gs_pz = {zeros(1, length(ws))};

for i = 1:length(ws)
  w = ws(i);
  Gs_pz{i} = linearize(mdl, io, 0.1);
  Gs_pz{i}.InputName  = {'Fu', 'Fv'};
  Gs_pz{i}.OutputName = {'Du', 'Dv'};
end

freqs = logspace(2, 3, 1000);

figure;
ax1 = subplot(2,1,1);
hold on;
for i = 1:length(ws)
  plot(freqs, abs(squeeze(freqresp(Gs_pz{i}(1, 1), freqs, 'Hz'))));
end
hold off;
xlim([freqs(1), freqs(end)]);
set(gca, 'XScale', 'log'); set(gca, 'YScale', 'log');
set(gca, 'XTickLabel',[]);
ylabel('Magnitude [m/N]');

ax2 = subplot(2,1,2);
hold on;
for i = 1:length(ws)
  plot(freqs, 180/pi*angle(squeeze(freqresp(Gs_pz{i}(1, 1), freqs, 'Hz'))), 'DisplayName', sprintf('w = %.0f [rpm]', ws(i)*60/2/pi));
end
hold off;
yticks(-180:90:180);
ylim([-180 180]);
xlim([freqs(1), freqs(end)]);
set(gca, 'XScale', 'log'); set(gca, 'YScale', 'lin');
xlabel('Frequency [Hz]'); ylabel('Phase [deg]');
legend('Location', 'northeast');
linkaxes([ax1,ax2],'x');

% Controller design
% We design a controller based on the identification when the system is not rotating.

% The obtained controller is a lead-lag controller with the following transfer function.

Kll = 2.0698e09*(s+40.45)*(s+1.181)/((s+0.01)*(s+198.4)*(s+2790));
K = [Kll 0;
     0   Kll];

K.InputName  = {'Du', 'Dv'};
K.OutputName = {'Fu', 'Fv'};



% The loop gain is displayed figure [[fig:Gvc_loop_gain]].


freqs = logspace(-2, 2, 1000);

figure;
% Amplitude
ax1 = subplot(2,1,1);
hold on;
plot(freqs, abs(squeeze(freqresp(Gvc('Du', 'fu')*Kll, freqs, 'Hz'))), '-');
set(gca,'xscale','log'); set(gca,'yscale','log');
ylabel('Amplitude [m/N]');
set(gca, 'XTickLabel',[]);
hold off;
% Phase
ax2 = subplot(2,1,2);
hold on;
plot(freqs, 180/pi*unwrap(angle(squeeze(freqresp(Gvc('Du', 'fu')*Kll, freqs, 'Hz')))), '-');
set(gca,'xscale','log');
yticks(-180:180:180);
ylim([-180 180]);
xlabel('Frequency [Hz]'); ylabel('Phase [deg]');
hold off;
linkaxes([ax1,ax2],'x');

% Controlling the rotating system
% We here want to see if the system is robust with respect to the rotation speed.
% We use the controller that was designed based on the dynamics of the non-rotating system.

% Let's first plot the SISO loop gain.


freqs = logspace(-2, 2, 1000);

figure;
% Amplitude
ax1 = subplot(2,1,1);
hold on;
plot(freqs, abs(squeeze(freqresp(Gs_vc{1}(1, 1)*Kll, freqs, 'Hz'))), '-');
plot(freqs, abs(squeeze(freqresp(Gs_vc{2}(1, 1)*Kll, freqs, 'Hz'))), '-');
plot(freqs, abs(squeeze(freqresp(Gs_vc{5}(1, 1)*Kll, freqs, 'Hz'))), '-');
set(gca,'xscale','log'); set(gca,'yscale','log');
ylabel('Amplitude [m/N]');
set(gca, 'XTickLabel',[]);
hold off;

% Phase
ax2 = subplot(2,1,2);
hold on;
plot(freqs, 180/pi*angle(squeeze(freqresp(Gs_vc{1}(1, 1)*Kll, freqs, 'Hz'))), 'DisplayName', sprintf('%.0f rpm', ws(1)./(2*pi).*60));
plot(freqs, 180/pi*angle(squeeze(freqresp(Gs_vc{2}(1, 1)*Kll, freqs, 'Hz'))), 'DisplayName', sprintf('%.0f rpm', ws(2)./(2*pi).*60));
plot(freqs, 180/pi*angle(squeeze(freqresp(Gs_vc{5}(1, 1)*Kll, freqs, 'Hz'))), 'DisplayName', sprintf('%.0f rpm', ws(5)./(2*pi).*60));
set(gca,'xscale','log');
yticks(-180:180:180);
ylim([-180 180]);
xlabel('Frequency [Hz]'); ylabel('Phase [deg]');
hold off;
legend('Location', 'northeast');
linkaxes([ax1,ax2],'x');



% #+NAME: fig:loop_gain_turning
% #+CAPTION: Loop Gain $G_u * K$
% [[file:figs/loop_gain_turning.png]]

% We can now compute the close-loop systems.

Gvc_cl = {zeros(1, length(ws))};
for i = 1:length(ws)
  Gvc_cl{i} = feedback(Gs_vc{i}, K, 'name');
end



% Let's now look on figure [[fig:evolution_poles_u]] at the evolution of the poles of the system when closing only one loop (controlling only one direction). We see that two complex conjugate poles are approaching the real axis and then they separate: one goes to positive real part and the other goes to negative real part.
% The system then goes unstable at some point (here for $\omega=60rpm$).


figure;
hold on;
for i = 1:length(ws)
  sys = feedback(Gs_vc{i}(1, 1), K(1, 1), 'name');
  plot(real(pole(sys)), imag(pole(sys)), 'x', 'DisplayName', sprintf('$\\omega = %.0f rpm$', ws(i)/(2*pi)*60));
end
hold off;
xlim([-80, 10]);
xlabel('Real Axis'); ylabel('Imaginary Axis');
legend('Location', 'northeast');



% #+NAME: fig:evolution_poles_u
% #+CAPTION: Evolution of the poles of the closed-loop system when closing just one loop
% [[file:figs/evolution_poles_u.png]]

% If we look at the poles of the closed loop-system for multiple rotating speed (figure [[fig:poles_cl_system]]) when closing the two loops (MIMO system), we see that they all have a negative real part (stable system), and their evolution on the complex plane is rather small compare to the open loop evolution.


figure;
hold on;
for i = 1:length(ws)
  plot(real(pole(Gvc_cl{i})), imag(pole(Gvc_cl{i})), 'x', 'DisplayName', sprintf('$\\omega = %.0f rpm$', ws(i)/(2*pi)*60));
end
hold off;
xlim([-80, 0]);
xlabel('Real Axis'); ylabel('Imaginary Axis');
legend('Location', 'northeast');

% Close loop performance
% First, we create the closed loop systems. Then, we plot the transfer function from the reference signals $[r_u, r_v]$ to the output $[d_u, d_v]$ (figure [[fig:perfcomp]]).


freqs = logspace(-2, 3, 1000);

figure;
ax1 = subplot(1,2,1);
title('$G_{r_u \to d_u}$')
hold on;
for i = 1:length(ws)
  sys = Gvc_cl{i}*K;
  plot(freqs, abs(squeeze(freqresp(sys(1, 1), freqs, 'Hz'))));
end
hold off;
xlim([freqs(1), freqs(end)]);
ylim([1e-4, 10]);
set(gca, 'XScale', 'log'); set(gca, 'YScale', 'log');
xlabel('Frequency [Hz]'); ylabel('Magnitude [m/N]');

ax2 = subplot(1,2,2);
title('$G_{r_u \to d_v}$')
hold on;
for i = 1:length(ws)
  sys = Gvc_cl{i}*K;
  plot(freqs, abs(squeeze(freqresp(sys(1, 2), freqs, 'Hz'))), 'DisplayName', sprintf('$\\omega = %.0f rpm$', ws(i)/(2*pi)*60));
end
hold off;
xlim([freqs(1), freqs(end)]);
ylim([1e-4, 10]);
set(gca, 'XScale', 'log'); set(gca, 'YScale', 'log');
xlabel('Frequency [Hz]');
legend('Location', 'northeast')

linkaxes([ax1,ax2],'x');

% Campbell Diagram for the close loop system


m = mlight;
k = kvc;
c = 0.1*sqrt(k*m);

wsvc = linspace(0, 10, 100); % [rad/s]

polesvc = zeros(8, length(wsvc));

for i = 1:length(wsvc)
  Gs = (m*s^2 + (k-m*w^2))/((m*s^2 + (k - m*w^2))^2 + (2*m*w*s)^2);
  Gcs = (2*m*w*s)/((m*s^2 + (k - m*w^2))^2 + (2*m*w*s)^2);

  G = [Gs, Gcs; Gcs, Gs];
  G.InputName  = {'Fu', 'Fv'};
  G.OutputName = {'Du', 'Dv'};

  polei = pole(feedback(G, K, 'name'));
  polesvc(:, i) = sort(polei(imag(polei) > 0));

  polei = pole(feedback(G, 10*K, 'name'));
  polesvcb(:, i) = sort(polei(imag(polei) > 0));
end

figure;
% Amplitude
ax1 = subplot(1,2,1);
hold on;
for i = 1:8
  plot(wsvc, real(polesvc(i, :)), 'b')
  plot(wsvc, real(polesvcb(i, :)), 'r')
end
plot(wsvc, zeros(size(wsvc)), 'k--')
hold off;
xlabel('Rotation Frequency [rad/s]');
ylabel('Pole Real Part');

ax2 = subplot(1,2,2);
hold on;
for i = 1:8
  plot(wsvc, imag(polesvc(i, :)), 'b')
  plot(wsvc, imag(polesvcb(i, :)), 'r')
end
hold off;
xlabel('Rotation Frequency [rad/s]');
ylabel('Pole Imaginary Part');
